/*
 * Copyright (C) 2005-2012 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file check_wav.c Detect and delete a wav header. */

#include <regex.h>

#include "para.h"
#include "string.h"
#include "list.h"
#include "sched.h"
#include "buffer_tree.h"
#include "error.h"
#include "check_wav.h"

/** Length of a standard wav header. */
#define WAV_HEADER_LEN 44

enum check_wav_state {
	CWS_NEED_HEADER,
	CWS_HAVE_HEADER,
	CWS_NO_HEADER,
};

struct check_wav_context {
	enum check_wav_state state;
	struct btr_node *btrn;
	size_t min_iqs;
	/* Command line args. */
	struct wav_params params;
	/* Extracted from the wav header.*/
	unsigned channels;
	unsigned sample_format;
	unsigned sample_rate;
};

void check_wav_pre_select(struct sched *s, struct check_wav_context *cwc)
{
	int ret = btr_node_status(cwc->btrn, cwc->min_iqs, BTR_NT_INTERNAL);
	if (ret != 0)
		sched_min_delay(s);
}

static int check_wav_exec(struct btr_node *btrn, const char *cmd, char **result)
{
	struct check_wav_context *cwc = btr_context(btrn);
	int val, header_val, given, arg;

	header_val = cwc->channels;
	arg = cwc->params.channels_arg;
	given = cwc->params.channels_given;
	if (!strcmp(cmd, "channels"))
		goto out;

	header_val = cwc->sample_rate;
	arg = cwc->params.sample_rate_arg;
	given = cwc->params.sample_rate_given;
	if (!strcmp(cmd, "sample_rate"))
		goto out;

	header_val = cwc->sample_format;
	arg = cwc->params.sample_format_arg;
	given = cwc->params.sample_format_given;
	if (!strcmp(cmd, "sample_format"))
		goto out;

	return -ERRNO_TO_PARA_ERROR(ENOTSUP);
out:
	if (given)
		val = arg;
	else {
		switch (cwc->state) {
		case CWS_HAVE_HEADER:
			val = header_val;
			break;
		case CWS_NO_HEADER:
			/*
			 * No wav header available and no value specified at
			 * the command line. Maybe one of our parent nodes
			 * knows.
			 */
			if (btr_exec_up(btr_parent(cwc->btrn), cmd, result) >= 0)
				return 1;
			/* Use default value */
			val = arg;
			break;
		default:
			return -E_BTR_NAVAIL;
		}
	}
	*result = make_message("%d", val);
	return 1;
}

int check_wav_post_select(struct check_wav_context *cwc)
{
	struct btr_node *btrn = cwc->btrn;
	unsigned char *a;
	size_t sz;
	int ret;
	uint16_t bps; /* bits per sample */
	const char *sample_formats[] = {SAMPLE_FORMATS};

	if (!btrn)
		return 0;
	ret = btr_node_status(btrn, cwc->min_iqs, BTR_NT_INTERNAL);
	if (ret <= 0)
		goto out;
	if (cwc->state != CWS_NEED_HEADER)
		goto pushdown;
	btr_merge(btrn, cwc->min_iqs);
	sz = btr_next_buffer(btrn, (char **)&a);
	if (sz < cwc->min_iqs) /* file size less than WAV_HEADER_SIZE */
		goto pushdown;
	cwc->min_iqs = 0;
	/*
	 * The default byte ordering assumed for WAVE data files is
	 * little-endian. Files written using the big-endian byte ordering
	 * scheme have the identifier RIFX instead of RIFF.
	 */
	if (a[0] != 'R' || a[1] != 'I' || a[2] != 'F' ||
			(a[3] != 'F' && a[3] != 'X')) {
		PARA_NOTICE_LOG("wav header not found\n");
		cwc->state = CWS_NO_HEADER;
		goto out;
	}
	PARA_INFO_LOG("found wav header\n");
	cwc->state = CWS_HAVE_HEADER;
	/* Only set those values which have not already been set. */
	cwc->channels = (unsigned)a[22];
	cwc->sample_rate = a[24] + (a[25] << 8) + (a[26] << 16) + (a[27] << 24);
	bps = a[34] + ((unsigned)a[35] << 8);
	if (bps != 8 && bps != 16) {
		PARA_WARNING_LOG("%u bps not supported, assuming 16\n",
			bps);
		bps = 16;
	}
	/*
	 * 8-bit samples are stored as unsigned bytes, ranging from 0
	 * to 255.  16-bit samples are stored as 2's-complement signed
	 * integers, ranging from -32768 to 32767.
	 */
	if (bps == 8)
		cwc->sample_format = SF_U8;
	else
		cwc->sample_format = (a[3] == 'F')?
			SF_S16_LE : SF_S16_BE;
	PARA_NOTICE_LOG("%dHz, %s, %s\n", cwc->sample_rate,
		cwc->channels == 1? "mono" : "stereo",
		sample_formats[cwc->sample_format]);
	btr_consume(btrn, WAV_HEADER_LEN);
pushdown:
	btr_pushdown(btrn);
out:
	if (ret < 0)
		btr_remove_node(&cwc->btrn);
	return ret;
}

struct check_wav_context *check_wav_init(struct btr_node *parent,
		struct btr_node *child, struct wav_params *params,
		struct btr_node **cw_btrn)
{
	struct check_wav_context *cwc = para_calloc(sizeof(*cwc));

	cwc->state = CWS_NEED_HEADER;
	cwc->min_iqs = WAV_HEADER_LEN;
	cwc->params = *params;
	cwc->btrn = btr_new_node(&(struct btr_node_description)
		EMBRACE(.name = "check_wav", .parent = parent, .child = child,
		.handler = check_wav_exec, .context = cwc));
	if (cw_btrn)
		*cw_btrn = cwc->btrn;
	return cwc;
}

void check_wav_shutdown(struct check_wav_context *cwc)
{
	free(cwc);
}
