/*
 * Copyright (C) 2006-2012 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file file_write.c simple output plugin for testing purposes */

#include <regex.h>
#include <sys/types.h>
#include <sys/time.h>

#include "para.h"
#include "list.h"
#include "sched.h"
#include "ggo.h"
#include "buffer_tree.h"
#include "write.h"
#include "write_common.h"
#include "string.h"
#include "fd.h"
#include "file_write.cmdline.h"
#include "error.h"

/** Data specific to the file writer. */
struct private_file_write_data {
	/** The file descriptor of the output file. */
	int fd;
};

/*
 * Get a random filename.
 *
 * This is by no means a secure way to create temporary files in a hostile
 * directory like \p /tmp. However, we use it only for creating temp files in
 * ~/.paraslash, for which it is OK. Result must be freed by the caller.
 */
__must_check __malloc static char *random_filename(void)
{
	char *result, *home = para_homedir();
	struct timeval tv;

	gettimeofday(&tv, NULL);
	srandom(tv.tv_usec);
	result = make_message("%s/.paraslash/%08lu", home,
		para_random(99999999));
	free(home);
	return result;
}

static int prepare_output_file(struct writer_node *wn)
{
	struct file_write_args_info *conf = wn->conf;
	char *filename;
	int ret;
	struct private_file_write_data *pfwd = para_calloc(sizeof(*pfwd));

	if (conf->filename_given)
		filename = conf->filename_arg;
	else
		filename = random_filename();
	ret = para_open(filename, O_WRONLY | O_CREAT, S_IRUSR | S_IWUSR);
	if (!conf->filename_given)
		free(filename);
	if (ret < 0)
		goto out;
	pfwd->fd = ret;
	ret = mark_fd_blocking(pfwd->fd);
	if (ret < 0)
		goto out_close;
	wn->private_data = pfwd;
	return 1;
out_close:
	close(pfwd->fd);
out:
	free(pfwd);
	return ret;
}

static void file_write_pre_select(struct sched *s, struct task *t)
{
	struct writer_node *wn = container_of(t, struct writer_node, task);
	struct private_file_write_data *pfwd = wn->private_data;
	int ret = btr_node_status(wn->btrn, wn->min_iqs, BTR_NT_LEAF);

	if (ret == 0)
		return;
	if (ret < 0 || !pfwd)
		return sched_min_delay(s);
	para_fd_set(pfwd->fd, &s->wfds, &s->max_fileno);
}

static void file_write_close(struct writer_node *wn)
{
	struct private_file_write_data *pfwd = wn->private_data;

	if (!pfwd)
		return;
	close(pfwd->fd);
	free(pfwd);
}

static void file_write_post_select(__a_unused struct sched *s,
		struct task *t)
{
	struct writer_node *wn = container_of(t, struct writer_node, task);
	struct private_file_write_data *pfwd = wn->private_data;
	struct btr_node *btrn = wn->btrn;
	int ret;
	char *buf;
	size_t bytes;

	ret = task_get_notification(t);
	if (ret < 0)
		goto out;
	ret = btr_node_status(btrn, wn->min_iqs, BTR_NT_LEAF);
	if (ret <= 0)
		goto out;
	if (!pfwd) {
		ret = prepare_output_file(wn);
		goto out;
	}
	if (!FD_ISSET(pfwd->fd, &s->wfds))
		return;
	bytes = btr_next_buffer(btrn, &buf);
	assert(bytes > 0);
	//PARA_INFO_LOG("writing %zu\n", bytes);
	ret = write(pfwd->fd, buf, bytes);
	if (ret < 0)
		goto out;
	btr_consume(btrn, ret);
out:
	if (ret < 0)
		btr_remove_node(&wn->btrn);
	t->error = ret;
}

__malloc static void *file_write_parse_config_or_die(int argc, char **argv)
{
	struct file_write_args_info *conf = para_calloc(sizeof(*conf));

	/* exits on errors */
	file_write_cmdline_parser(argc, argv, conf);
	return conf;
}

static void file_write_free_config(void *conf)
{
	file_write_cmdline_parser_free(conf);
}

/** the init function of the file writer */
void file_write_init(struct writer *w)
{
	struct file_write_args_info dummy;

	file_write_cmdline_parser_init(&dummy);
	w->pre_select = file_write_pre_select;
	w->post_select = file_write_post_select;
	w->parse_config_or_die = file_write_parse_config_or_die;
	w->free_config = file_write_free_config;
	w->close = file_write_close;
	w->help = (struct ggo_help) {
		.short_help = file_write_args_info_help,
		.detailed_help = file_write_args_info_detailed_help
	};
	file_write_cmdline_parser_free(&dummy);
}
