/*
 * Copyright (C) 2005-2012 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file filter_common.c Common helper functions for filter input/output. */

#include <regex.h>
#include <sys/types.h>

#include "para.h"
#include "list.h"
#include "sched.h"
#include "fd.h"
#include "ggo.h"
#include "buffer_tree.h"
#include "filter.h"
#include "error.h"
#include "string.h"

/** The array of supported filters. */
struct filter filters[NUM_SUPPORTED_FILTERS] = {FILTER_ARRAY};

/**
 * Call the init function of each supported filter.
 * \sa filter::init
 */
void filter_init(void)
{
	int i;

	FOR_EACH_SUPPORTED_FILTER(i)
		filters[i].init(filters + i);
}

/*
 * If the filter has a command line parser and options is not NULL, run it.
 * Returns filter_num on success, negative on errors
 */
static int parse_filter_args(int filter_num, char *options, void **conf)
{
	struct filter *f = &filters[filter_num];
	int ret, argc;
	char **argv;

	if (!f->parse_config)
		return strlen(options)? -E_BAD_FILTER_OPTIONS : filter_num;
	argc = create_shifted_argv(options, " \t", &argv);
	if (argc < 0)
		return -E_BAD_FILTER_OPTIONS;
	argv[0] = para_strdup(f->name);
	ret = f->parse_config(argc, argv, conf);
	free_argv(argv);
	return ret < 0? ret : filter_num;
}

/**
 * Check the filter command line options.
 *
 * \param fa The command line options.
 * \param conf Points to the filter configuration upon successful return.
 *
 * Check if \a fa starts with a the name of a supported filter, followed by
 * a colon. If yes, call the command line parser of that filter.
 *
 * \return On success, the number of the filter is returned and \a conf
 * is initialized to point to the filter configuration determined by \a fa.
 * On errors, a negative value is returned.
 *
 * Note: If \a fa specifies a filter that has no command line parser success is
 * returned, and \a conf is initialized to \p NULL.
 *
 * \sa filter::parse_config
 */
int check_filter_arg(char *fa, void **conf)
{
	int j;

	*conf = NULL;
//	PARA_DEBUG_LOG("arg: %s\n", fa);
	FOR_EACH_SUPPORTED_FILTER(j) {
		const char *name = filters[j].name;
		size_t len = strlen(name);
		char c;
		if (strlen(fa) < len)
			continue;
		if (strncmp(name, fa, len))
			continue;
		c = fa[len];
		if (c && c != ' ')
			continue;
		if (c && !filters[j].parse_config)
			return -E_BAD_FILTER_OPTIONS;
		return parse_filter_args(j, c? fa + len + 1 :
			fa + strlen(fa), conf);
	}
	return -E_UNSUPPORTED_FILTER;
}

/**
 * Print help text of each filter to stdout.
 *
 * \param detailed If non-zero, print detailed help.
 */
void print_filter_helps(int detailed)
{
	int i;

	printf_or_die("\nAvailable filters: \n\t");
	FOR_EACH_SUPPORTED_FILTER(i)
		printf_or_die("%s%s", i? " " : "", filters[i].name);
	printf_or_die("\n\n");

	FOR_EACH_SUPPORTED_FILTER(i) {
		struct filter *f = filters + i;

		if (!f->help.short_help)
			continue;
		printf_or_die("Options for %s:\n", f->name);
		ggo_print_help(&f->help, detailed);
	}
}

/**
 * Set select timeout of the the scheduler.
 *
 * \param s The scheduler.
 * \param t The task struct of this filter.
 *
 * This looks at the status of the btr node of the filter. If data is available
 * in the input queue of the filter, or if an error occured, a minimal timeout
 * for the next select call is requested from the scheduler. Otherwise the
 * scheduler timeout is left unchanged.
 */
void generic_filter_pre_select(struct sched *s, struct task *t)
{
	struct filter_node *fn = container_of(t, struct filter_node, task);

	t->error = 0;
	if (btr_node_status(fn->btrn, fn->min_iqs, BTR_NT_INTERNAL) != 0)
		sched_min_delay(s);
}

#ifdef WORDS_BIGENDIAN
#define DECODER_SAMPLE_FORMAT SF_S16_BE
#else
#define DECODER_SAMPLE_FORMAT SF_S16_LE
#endif

/**
 * Execute a btr command for a decoder.
 *
 * The buffer tree nodes of the writers ask the parent nodes about sample_rate,
 * channels count and sample format. This function is called by all decoders to
 * answer these queries.
 *
 * \param cmd The command to be executed by the child node.
 * \param sample_rate Known to the decoder.
 * \param channels Known to the decoder.
 * \param result Ascii representation on the answer is stored here.
 *
 * \return Standard.
 */
int decoder_execute(const char *cmd, unsigned sample_rate, unsigned channels,
		char **result)
{
	if (!strcmp(cmd, "sample_rate")) {
		if (sample_rate == 0)
			return -E_BTR_NAVAIL;
		*result = make_message("%u", sample_rate);
		return 1;
	}
	if (!strcmp(cmd, "channels")) {
		if (channels == 0)
			return -E_BTR_NAVAIL;
		*result = make_message("%u", channels);
		return 1;
	}
	if (!strcmp(cmd, "sample_format")) {
		*result = make_message("%u", DECODER_SAMPLE_FORMAT);
		return 1;
	}
	return -ERRNO_TO_PARA_ERROR(ENOTSUP);
}
